import asyncio

from grpc.aio import server
from grpc_reflection.v1alpha import reflection as grpc_reflection

from .service import FastGRPCService


class FastGRPC:
    def __init__(self, *services: FastGRPCService, port: int = 50051, reflection: bool = False):
        self._server = server()
        self._server.add_insecure_port(f"[::]:{port}")

        for service in services:
            self.add_service(service)

        if reflection:
            service_names = [svc.get_service_name() for svc in services]
            service_names.append(grpc_reflection.SERVICE_NAME)
            grpc_reflection.enable_server_reflection(service_names, self._server)

    def add_service(self, service: FastGRPCService):
        register_function = getattr(service.pb2_grpc, f"add_{service.name}Servicer_to_server")
        register_function(service, self._server)

    def run(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.run_async())
        loop.close()

    async def run_async(self):
        await self._server.start()
        await self._server.wait_for_termination()



import pathlib

import jinja2
from grpc_tools import protoc

from .models import Service


TEMPLATE_DIR_PATH = pathlib.Path(__file__).parent / "templates"
JINJA_ENV = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATE_DIR_PATH))



def render_proto(service: Service, proto_path: pathlib.Path):
    template = JINJA_ENV.get_template("service.proto")
    content = template.render(service=service)
    proto_path.mkdir(parents=True, exist_ok=True)
    proto_file_path = proto_path / f"{service.name.lower()}.proto"
    proto_file_path.write_text(content)


def compile_proto(service: Service, proto_path: pathlib.Path, grpc_path: pathlib.Path):
    grpc_path.mkdir(parents=True, exist_ok=True)
    protoc_args = [
        f"--proto_path={proto_path}",
        f"--python_out={grpc_path}",
        f"--grpc_python_out={grpc_path}",
        str(proto_path / f"{service.name.lower()}.proto"),
        f"--proto_path={proto_path}",
    ]
    status_code = protoc.main(protoc_args)

    if status_code != 0:
        raise Exception("Protobuf compilation failed")


def delete_proto(service: Service, proto_path: pathlib.Path):
    proto_file_path = proto_path / f"{service.name.lower()}.proto"
    proto_file_path.unlink(missing_ok=True)



import inspect
from types import NoneType, UnionType
from typing import Type, Union, get_origin

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from .enums import TYPE_MAPPING
from .models import Message, Field


def parse_type_sequence(name: str, python_type, args: list) -> Field:
    if len(args) != 1:
        raise TypeError(
            f"Field '{name}': type '{python_type}' must have only one subtype, not {len(args)}.",
        )

    inside_python_type = args[0]
    if (origin := get_origin(inside_python_type)) is not None:
        inside_args = list(inside_python_type.__args__)
        if origin in (Union, UnionType):
            return parse_type_union(name, inside_python_type, inside_args)
        else:
            raise TypeError(
                f"Field '{name}': unsupported type '{origin}' in '{inside_python_type}.",
            )
    if inside_python_type in TYPE_MAPPING:
        grpc_type = TYPE_MAPPING[inside_python_type].value
    elif inspect.isclass(inside_python_type) and issubclass(inside_python_type, BaseModel):
        grpc_type = inside_python_type.__name__
    else:
        raise TypeError(
            f"Field '{name}': unsupported type '{inside_python_type}' in type '{python_type}'.",
        )

    return Field(name=name, type=grpc_type, repeated=True)


def parse_type_mapping(name: str, python_type, args: list) -> Field:
    if len(args) != 2:
        raise TypeError(
            f"Field '{name}': type '{python_type}' must have two subtype, not {len(args)}.",
        )

    grpc_type = TYPE_MAPPING[dict].value
    python_map_key, python_map_value = args
    map_key, map_value = None, None
    if python_map_key in TYPE_MAPPING:
        map_key = TYPE_MAPPING[python_map_key].value
    else:
        raise TypeError(
            f"Field '{name}': unsupported type '{python_map_key}' in '{python_type}'.",
        )
    if python_map_value in TYPE_MAPPING:
        map_value = TYPE_MAPPING[python_map_value].value
    elif inspect.isclass(python_map_value) and issubclass(python_map_value, BaseModel):
        map_value = python_map_value.__name__
    else:
        raise TypeError(
            f"Field '{name}': unsupported type '{python_map_value}' in '{python_type}'.",
        )

    return Field(
        name=name,
        type=grpc_type,
        map_key=map_key,
        map_value=map_value,
    )


def parse_type_union(name: str, python_type, args: list) -> Field:
    if NoneType in args:
        args.remove(NoneType)
    if len(args) != 1:
        raise TypeError(
            f"Field '{name}': type '{python_type}' must have only one subtype , not {len(args)}. "
            "Tip: None/Optional type ignoring."
        )

    inside_python_type = args[0]
    if (origin := get_origin(inside_python_type)) is not None:
        inside_args = list(inside_python_type.__args__)
        if origin in (Union, UnionType):
            return parse_type_union(name, inside_python_type, inside_args)
    if inside_python_type in TYPE_MAPPING:
        grpc_type = TYPE_MAPPING[inside_python_type].value
    elif inspect.isclass(inside_python_type) and issubclass(inside_python_type, BaseModel):
        grpc_type = inside_python_type.__name__
    else:
        raise TypeError(
            f"Field '{name}': unsupported type '{inside_python_type}' in {python_type}'.",
        )

    return Field(name=name, type=grpc_type)


def parse_field(name: str, field: FieldInfo) -> Field:
    repeated = False
    map_key, map_value = None, None
    python_type = field.annotation
    if (origin := get_origin(python_type)) is not None:
        args = list(python_type.__args__)
        if origin in (Union, UnionType):
            return parse_type_union(name, python_type, args)
        if issubclass(origin, (list, tuple, set, frozenset)):
            return parse_type_sequence(name, python_type, args)
        elif issubclass(origin, dict):
            return parse_type_mapping(name, python_type, args)

    if python_type in TYPE_MAPPING:
        grpc_type = TYPE_MAPPING[python_type].value
    else:
        raise TypeError(f"Field '{name}': unsupported type '{python_type}'.")

    return Field(
        name=name,
        repeated=repeated,
        type=grpc_type,
        map_key=map_key,
        map_value=map_value,
    )


def get_message_from_model(model: Type[BaseModel]) -> Message:
    fields = {}

    for name, field in model.__fields__.items():
        fields[name] = parse_field(name=name, field=field)

    return Message(name=model.__name__, fields=fields)


def gather_models(model: Type[BaseModel]) -> set[Type[BaseModel]]:
    models = set()
    stack = [model]
    processed = set()

    while stack:
        model = stack.pop()
        models.add(model)
        processed.add(model)
        
        for field in model.__fields__.values():
            arg_stack = [field.annotation]
            while arg_stack:
                arg = arg_stack.pop()
                if get_origin(arg) is not None:
                    arg_stack.extend(arg.__args__)
                elif (
                        inspect.isclass(arg) and
                        issubclass(arg, BaseModel) and
                        arg not in processed and
                        arg not in stack
                ):
                    stack.append(arg)

    return models



import inspect
import pathlib
import sys
from typing import Any, Callable, Type

import grpc
from google.protobuf.json_format import MessageToDict, ParseDict
from pydantic import BaseModel

from . import proto


class grpc_method:
    def __init__(
            self,
            name: str | None = None,
            request_model: Type[BaseModel] | None = None,
            response_model: Type[BaseModel] | None = None,
            disable: bool = False,
    ):
        self._name = name
        self._request_model = request_model
        self._response_model = response_model
        self._disable = disable

    def __call__(self, function: Callable):
        async def wrapper(self, request, context):
            request_model = wrapper._grpc_method_request_model
            response_model = wrapper._grpc_method_response_model

            request = request_model.model_validate(request, from_attributes=True)

            args = {"self": self, "request": request}
            if "context" in signature.parameters:
                args["context"] = context

            response = await function(**args)

            grpc_model = getattr(self.pb2, response_model.__name__)
            return ParseDict(
                response.model_dump(mode="json"),
                grpc_model(),
                ignore_unknown_fields=True,
            )

        signature = inspect.signature(function)
        wrapper.__name__ = function.__name__
        wrapper._grpc_method_enabled = not self._disable
        wrapper._grpc_method_name = self._name or function.__name__
        wrapper._grpc_method_request_model = (
            self._request_model or self.get_request_model_from_signature(signature)
        )
        wrapper._grpc_method_response_model = (
            self._response_model or self.get_response_model_from_signature(signature)
        )

        return wrapper

    @staticmethod
    def get_request_model_from_signature(signature) -> Type[BaseModel]:
        if "request" not in signature.parameters:
            raise TypeError("GRPC method should have 'request' parameter")
        request_parameter = signature.parameters["request"]
        if not issubclass(request_parameter.annotation, BaseModel):
            raise TypeError("GRPC method parameter 'request' should be pydantic model")
        return request_parameter.annotation

    @staticmethod
    def get_response_model_from_signature(signature) -> Type[BaseModel]:
        if not issubclass(signature.return_annotation, BaseModel):
            raise TypeError("GRPC method should have pydantic model in return annotation")
        return signature.return_annotation


class FastGRPCServiceMeta(type):
    def __init__(cls, name: str, bases: tuple[type], attributes: dict[str, Any]):
        super().__init__(name, bases, attributes)

        if not attributes.get("is_proxy", False):
            cls.name = attributes.pop("name", name)
            cls.proto_path = attributes.pop("proto_path", pathlib.Path.cwd())
            cls.grpc_path = attributes.pop("grpc_path", pathlib.Path.cwd())
            cls._methods = {}

            cls._setup()

    def _setup(cls):
        service = cls.build_proto_service()
        try:
            proto.render_proto(service=service, proto_path=cls.proto_path)
            proto.compile_proto(service=service, proto_path=cls.proto_path,
                                grpc_path=cls.grpc_path)
        finally:
            proto.delete_proto(service=service, proto_path=cls.proto_path)

        grpc_path = str(cls.grpc_path)
        if grpc_path not in sys.path:
            sys.path.append(grpc_path)
        cls.pb2 = __import__(f"{service.name.lower()}_pb2")
        cls.pb2_grpc = __import__(f"{service.name.lower()}_pb2_grpc")
        cls.Client = cls.get_client()

    def build_proto_service(cls) -> proto.Service:
        cls._methods.clear()
        methods = {}
        messages = {}
        models = set()
        for attribute_name in dir(cls):
            attribute = getattr(cls, attribute_name)
            if not getattr(attribute, "_grpc_method_enabled", False):
                continue

            method_name = attribute._grpc_method_name
            models |= proto.gather_models(attribute._grpc_method_request_model)
            models |= proto.gather_models(attribute._grpc_method_response_model)
            request_message = proto.get_message_from_model(attribute._grpc_method_request_model)
            response_message = proto.get_message_from_model(attribute._grpc_method_response_model) 

            cls._methods[method_name] = attribute

            methods[method_name] = proto.Method(
                name=method_name,
                request=request_message,
                response=response_message,
            )
            for model in models:
                message = proto.get_message_from_model(model)
                messages[message.name] = message

        return proto.Service(name=cls.name, methods=methods, messages=messages)

    def get_client(cls) -> type:
        class_name = f"{cls.name}Client"
        attributes = {}
        for grpc_method_name, method in cls._methods.items():
            request_message_class = getattr(cls.pb2, method._grpc_method_request_model.__name__)
            response_message_class = method._grpc_method_response_model

            async def wrapper(
                    self,
                    request,
                    _grpc_method_name=grpc_method_name,
                    _request_message_class=request_message_class,
                    _response_message_class=response_message_class,
            ):
                grpc_method = getattr(self._stub, _grpc_method_name)
                grpc_request = ParseDict(request.model_dump(mode="json"), _request_message_class())
                response = await grpc_method(request=grpc_request)
                response_dict = MessageToDict(response)
                return _response_message_class.model_validate(response_dict)

            attributes[method.__name__] = wrapper

        def __init__(self, host: str, port: int):
            channel = grpc.aio.insecure_channel(f"{host}:{port}")
            self._stub = getattr(cls.pb2_grpc, f"{cls.name}Stub")(channel)

        attributes["__init__"] = __init__

        return type(class_name, (), attributes)

    
class FastGRPCService(metaclass=FastGRPCServiceMeta):
    is_proxy: bool = True

    def __getattribute__(self, __name: str) -> Any:
        methods = object.__getattribute__(self, "_methods")
        if __name in methods:
            return methods[__name].__get__(self, self.__class__)
        return object.__getattribute__(self, __name)

    def get_service_name(self) -> str:
        return self.pb2.DESCRIPTOR.services_by_name[self.name].full_name



from pydantic import BaseModel, model_validator

from .enums import TypeEnum


class Field(BaseModel):
    name: str
    type: str
    repeated: bool = False
    map_key: str | None = None
    map_value: str | None = None

    @model_validator(mode="after")
    def root_validator(self):
        if self.type is TypeEnum.MAP:
            if self.repeated:
                raise ValueError("Field 'repeated' cannot be True, when protobuf type is 'map'")
            if self.map_key is None:
                raise ValueError("Field 'map_key' must be set, when protobuf type is 'map'")
            if self.map_value is None:
                raise ValueError("Field 'map_value' must be set, when protobuf type is 'map'")

        return self


class Message(BaseModel):
    name: str
    fields: dict[str, Field]


class Method(BaseModel):
    name: str
    request: Message
    response: Message


class Service(BaseModel):
    name: str
    methods: dict[str, Method]
    messages: dict[str, Message]



import enum


class StatusCode(enum.IntEnum):
    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    UNIMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16



import enum
import uuid

from pydantic import EmailStr, NegativeInt, NonNegativeInt, NonPositiveInt, PositiveInt


class TypeEnum(str, enum.Enum):
    DOUBLE = "double"
    FLOAT = "float"
    INT32 = "int32"
    INT64 = "int64"
    UINT32 = "uint32"
    UINT64 = "uint64"
    SINT32 = "sint32"
    SINT64 = "sint64"
    FIXED32 = "fixed32"
    FIXED64 = "fixed64"
    SFIXED32 = "sfixed32"
    SFIXED64 = "sfixed64"
    BOOL = "bool"
    STRING = "string"
    BYTES = "bytes"
    MAP = "map"


TYPE_MAPPING = {
    int: TypeEnum.INT64,
    NonPositiveInt: TypeEnum.INT64,
    NegativeInt: TypeEnum.INT64,
    NonNegativeInt: TypeEnum.UINT64,
    PositiveInt: TypeEnum.UINT64,
    float: TypeEnum.FLOAT,
    bool: TypeEnum.BOOL,
    str: TypeEnum.STRING,
    EmailStr: TypeEnum.STRING,
    uuid.UUID: TypeEnum.STRING,
    bytes: TypeEnum.BYTES,
    dict: TypeEnum.MAP,
}
